#!/usr/bin/python
"""
    Gterminal pluggable gobject-io readline magic terminal like interface.
    Copyright (C) 2008  James Shubin, McGill University
    Written for McGill University by James Shubin <purpleidea@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
# TODO: which variables/functions should be _local and which should be __private ?
import os
import sys
import math
import atexit
import termios
import tty
import gobject
import gobject_once
import string
import curses.ascii
import signal
import errno
import readline		# TODO: add tab completion, etc...
# TODO: should this be replaced with glib.get_* functions?
# http://www.pygtk.org/docs/pygobject/glib-functions.html#function-glib--get-user-config-dir
import xdg.BaseDirectory
import tcrecipe

__all__ = ['gterminal']
DEBUG = True

class gterminal:
	"""class to encapsulate the magic readline functionality."""

	def __init__(self, name,
		ps1='%s> ' % os.path.splitext(os.path.basename(__file__))[0]):
		"""be able to build and customize your own gterminal object."""

		# MISC ########################################################
		self.name = name
		self.ps1 = ps1
		self.good = string.digits + string.letters + ' '

		# save initial terminal settings
		self.__termios = termios.tcgetattr(sys.stdin)
		self.__restored = False

		self.__messages = {}			# dictionary of messages

		# XDG #########################################################
		# uncomment below two lines for adding a config directory
		#self.__conf = xdg.BaseDirectory.load_first_config(self.name) or
		#xdg.BaseDirectory.save_config_path(self.name)
		self.__data = (
			list(xdg.BaseDirectory.load_data_paths(self.name)) or
			[xdg.BaseDirectory.save_data_path(self.name)]
		)[0]	# take the first element in the resultant list

		# HISTORY #####################################################
		histfile = os.path.join(self.__data, 'history')
		try: readline.read_history_file(histfile)
		except IOError: pass
		atexit.register(readline.write_history_file, histfile)

		# DISPLAY #####################################################
		self.__term = tcrecipe.TerminalController()
		self.__block = tcrecipe.Block(self.__term)
		self.__buffer = ''		# text to display on screen
		self.__latest = ''		# text buffer for last item
		self.__xpos = 0			# position of the cursor
		self.__ypos = self.__ypos_index_latest()	# command hx pos

		# GO GO GO... #################################################
		tty.setcbreak(sys.stdin, termios.TCSANOW)	# setup terminal
		gobject.io_add_watch(sys.stdin, gobject.IO_IN, self.__io_cb)
		#gobject.io_add_watch(sys.stdin, gobject.IO_HUP, self.__io_cb)
		self.__block.update(self.__prompt())	# display initial prompt
		atexit.register(self.restore)		# register exit handler!

		# OLD/TESTING #################################################
		# capture stdin
		#import fcntl
		#flags = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
		#fcntl.fcntl(sys.stdin, fcntl.F_SETFL, flags | os.O_NONBLOCK)


	# MISC ################################################################
	def restore(self):
		"""restore original terminal settings."""
		if not self.__restored:
			termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self.__termios)
			self.__restored = True	# only let this be run once


	# MESSAGING ###########################################################
	def __message_id(self):
		"""returns an available message id for use in a new message."""
		tmp = [0]		# list of id's that are not available
		m = self.__messages	# dictionary of messages
		while True:
			i = max([x for x in m.keys()+tmp if type(x) is int]) + 1
			if i in m:
				# not available
				tmp.append(i)
			else:
				return i


	def msg(self, message, lifetime=15, name=None):
		"""add a notification message above the prompt. if lifetime is
		None, then it never expires. this returns a message handle."""
		# when will it expire... (and dissapear)
		now = gobject.get_current_time()
		if lifetime is None: expiry = None
		else: expiry = int(math.floor(now + lifetime))

		if name is None:
			name = self.__message_id()
		# if message already exists, don't modify the now field.
		if name in self.__messages: now = self.__messages[name][2]

		self.__messages[name] = ( (message, expiry, now ) )

		gobject.timeout_add_seconds(0, self.__update)	# update now
		# and schedule an update for the expected expiry time
		if lifetime is not None:
			gobject.timeout_add_seconds(lifetime, self.__update)
		return name


	def msg_clear(self, handle):
		"""schedules the immediate clear of a particular message."""
		if handle in self.__messages:
			del self.__messages[handle]
			gobject.timeout_add_seconds(0, self.__update)


	# DISPLAY #############################################################
	def __prompt(self, buf=''):
		"""the prompt."""
		tmp = '${GREEN}%s${NORMAL}%s' % (self.ps1, buf)
		return self.__term.render(tmp)


	def __update(self):
		"""updates/refreshes the prompt. does message scheduling too."""
		# filter out old messages by dictionary comprehension
		self.__messages = dict([ (key,(msg,exp,now)) for
		key,(msg,exp,now) in self.__messages.items() if
		exp > gobject.get_current_time() or exp is None ])
		lines = []
		# to display sensibly, sort by time added (the `now' field)
		# we want to see the *newest* items at the bottom near the >
		items = self.__messages.items()
		# [1] is the value, [2] is the now field (the third one)
		items.sort(key=lambda x: x[1][2])
		for (name, (msg, exp, now)) in items:
			line = '${BLUE}@${NORMAL} %s' % msg
			lines.append(self.__term.render(line))

		lines.append(self.__prompt(self.__buffer))
		self.__block.clear()
		self.__block.update(lines)

		# lastly, position the cursor where we expect it to be
		for i in range(len(self.__buffer) - self.__xpos):
			sys.stdout.write(self.__term.LEFT)
		sys.stdout.flush()
		return False


	def __safe(self, text=None, force=False):
		"""adds protector lines to avoid any messages squashing output.
		the force option should be used if have previously printed text
		manually, and you require the necessary padding to be added."""
		# this should only be called once per __update() call. multiple
		# calls interweaved with __update() cause it to seem as if many
		# commands have passed. not generally recommended for use ever.
		def pad():
			"""simple padding function required to protect text."""
			for i in self.__messages: print ''

		if text is not None:
			print text
			# this adds newlines so that any above
			# run text doesn't get munched over by
			# the messages that get printed. print
			# out one line for every message text.
			pad()
		elif force: pad()


	# IO ##################################################################
	def __getch(self, source):
		"""this should be called only when you either expect some data
		from the source, or are happy to have it block until some data
		arrives. it returns an array of numeric codes to process, but
		usually you'll just get one element. we only return the ord
		because for some possibilities (eg: >=256, the chr function
		can't handle the result and) we don't want to print an up arrow
		character anyways (whatever it might be). this func. uses yield
		magic, and the only way key presses can cause gobject to really
		block and wait for release, is by repeated <esc> pressing."""
		# NOTE: this was a *very* tricky/hard function to write. don't
		# change *anything* unless you know exactly what you're doing!
		SPECIAL_ORD = 91
		SPECIAL_MAP = {
			65: curses.KEY_UP, 66: curses.KEY_DOWN,
			67: curses.KEY_RIGHT, 68: curses.KEY_LEFT,
			51: curses.KEY_DC
		}
		sig = False	# previously named `esc'. now named sig.
		special = False
		munch = False
		while True:
			# *** capturing section ***
			if sig:
				signal.signal(signal.SIGALRM, lambda x,y: None)
				# sends '\0' on alarm
				signal.set_wakeup_fd(source.fileno())
				# FIXME: use ualarm to make time less than 1 sec
				signal.alarm(1)
			try:
				# it is very important to first set an initial
				# value for c, because if an error occurs when
				# trying to read, then the c will retain a bad
				# value (the last one it succesfully got), and
				# this looks like repeated keypresses. usually
				# it will look like repeated presses to escape
				c = '\0'	# this line is very important!
				c = source.read(1)
			except IOError, e:
				if errno.EINTR == e.errno: pass
				else: raise e
			finally:
				if sig: signal.alarm(0)	# disable the alarm
			o = ord(c)

			# *** processing section ***
			if munch:	# 1 char has been eaten. we can quit now
				assert o == munch, 'expected: chr(%d), got: chr(%d)' % (munch, o)
				break

			elif o == 0 and sig:
				# esc: last or singly pressed, yield and return
				yield curses.ascii.ESC
				break		# we're done, go return

			elif o == curses.ascii.ESC:
				special = False	# avoid sequence: <esc>[<esc>A
				# if `sig' was *already* true before we set it
				# below, then we've received one escape press,
				# followed by an escape, or arrow key press,
				# which hasn't been determined yet. (so loop)
				if sig:
					# sig: when pressed in rapid succession
					# the magic: let the caller get the esc
					# value now, so that we don't pause the
					# execution! this is the magic of yield
					yield curses.ascii.ESC
				sig = True

			# magic char after an esc
			elif sig and o == SPECIAL_ORD:
				# we have a special key?
				special = True
				sig = False

			elif special:
				if o in SPECIAL_MAP.keys():
					yield SPECIAL_MAP[o]
					# there seems to be an extra `126'
					# that hangs around after a delete
					# press, so gobble it out of the q
					if SPECIAL_MAP[o] == curses.KEY_DC:
						munch = 126
						sig = True	# to add timeout
					else:
						break
				else:
					if DEBUG:
						self.msg('DEBUG: special: chr(%d) not handled' % o)
					yield o
					break

			# if we get a null and sig is False, then the alarm
			# probably fired prematurely. it's probably fine to
			# return the null character and let the caller deal
			#elif o == 0:
			#	yield o
			#	break
			else:	# regular character
				yield o
				break
		return


	def __io_in(self, source):
		"""the io_in processor. broken up into this function to keep my
		code flat, and less nested. this should be called by __io_cb"""
		# NOTE: the ordering of the big if statement below is important!

		# loop in case there are multiple simultaneous presses
		for o in self.__getch(source):	# receives yields

			# avoid printing characters that we don't want to print
			if o not in [curses.ascii.ESC, curses.ascii.EOT]:
				try:
					temp = chr(o)
					# print out the character
					sys.stdout.write(temp)
					# flush is required to show it
					sys.stdout.flush()
				except ValueError:
					pass

			# NOTE: the ordering of this if statement is important!
			if o == curses.ascii.NL:	# 10: <enter> (run)
				if len(self.__buffer) > 0:
					readline.add_history(self.__buffer)
					# FIXME: eventually this will run some action instead of just printing
					# print out any text for display here.
					run = '${RED}run: ${NORMAL}%s' % self.__buffer
					print self.__term.render(run)

					# now, update the pos to be most recent
					self.__xpos = 0
					self.__ypos = self.__ypos_index_latest()
					self.__buffer = ''	# now erase it
				else:
					pass

				self.__safe(force=True)		# protect

			elif o == curses.ascii.TAB:	# 9: <tab> (complete)
				# TODO: implement some sort of tab completion
				pass

			elif o == curses.ascii.FF:	# 12: ^L (clear)
				sys.stdout.write(self.__term.CLEAR_SCREEN)

			elif o == curses.ascii.EOT:	# 4: ^D (end of text)
				self.__safe('exit')
				sys.exit(0)

			elif o == curses.ascii.ESC:	# 27: <esc>
				self.__safe('ESC')

			elif o == curses.KEY_LEFT:	# cursor: <prev>
				if not self.__cursor_prev():
					# beep if nowhere to move cursor
					sys.stdout.write(self.__term.BELL)

			elif o == curses.KEY_RIGHT:	# cursor: <next>
				if not self.__cursor_next():
					# beep if nowhere to move cursor
					sys.stdout.write(self.__term.BELL)

			elif o == curses.KEY_UP:	# hx: <prev>
				if not self.__history_prev():
					# beep if no past history to browse
					sys.stdout.write(self.__term.BELL)

			elif o == curses.KEY_DOWN:	# hx: <next>
				if not self.__history_next():
					# beep if no forward history to browse
					sys.stdout.write(self.__term.BELL)

			elif o == 64:	# quick hack to see HX	chr(64) == '@'
				print 'HISTORY'
				# TODO: run this code in an actual `history' cmd
				# once we've figured out the run specifications.
				self.__safe(self.__history(last=10))

			elif o == curses.KEY_DC:	# <delete>
				if not self.__cursor_del():
					# beep if we're out of buffer to remove
					sys.stdout.write(self.__term.BELL)

			# FIXME: which is it? 127 or 8 or both?
			elif (o == curses.ascii.DEL or	# 127: <backspace>
			o == curses.ascii.BS):		# 8: (on sam's computer)
				if not self.__cursor_bs():
					# beep if we're out of buffer to remove
					sys.stdout.write(self.__term.BELL)

			else:
				if not self.__cursor_add(o):
					# beep on `bad' chars
					if DEBUG: self.msg('DEBUG: unknown: chr(%d) not handled' % o)
					sys.stdout.write(self.__term.BELL)

			# XXX: maybe add a self.__safe() here, and allow people
			# to print whatever they like above. either that or add
			# all text to a common buffer and then print it all here

			self.__update()			# refresh the prompt

		return True


	def __io_cb(self, source, cb_condition):
		"""io handler."""
		if cb_condition == gobject.IO_IN:
			return self.__io_in(source)

		elif cb_condition == gobject.IO_HUP:
			# DOES THIS EVER HAPPEN? and
			# FIXME: what should we do if it does ?
			# TODO: maybe broadcast a gobject signal ?
			#log.warn('IO_HUP received.')
			return False
		else:
			#log.warn('unexpected cb_condition: %d' % cb_condition)
			return False


	# HISTORY #############################################################
	def __history(self, last=None):
		"""output the history to a nicely readable string format. you
		can specify a `last' integer to display only the last few."""
		l = readline.get_current_history_length()
		# add correct padding for history display
		pattern = '%' + str(len(str(l))) + 'd  %s'	# two spaces b/w
		output = []
		start = 1
		if last is not None: start = l+1-last
		for i in range(max(start, 1), l+1):
			txt = readline.get_history_item(i)
			output.append(pattern % (i, txt))

		return '\n'.join(output)


	def __ypos_index_latest(self):
		"""returns magic index of latest element."""
		return readline.get_current_history_length() + 1


	def __is_latest(self):
		"""returns whether ypos is referring to the last element."""
		return self.__ypos == self.__ypos_index_latest()


	def __history_replace(self):
		"""put all the magic of adding and removing characters from
		readline databases into one function that is easy to call."""
		# update any items except the latest one
		if not self.__is_latest():
			readline.replace_history_item(self.__ypos-1, self.__buffer)


	def __history_prev(self):
		"""command to cause previous history to be displayed."""
		if self.__ypos > 1:	# readline is 1 indexed
			# save latest
			if self.__is_latest(): self.__latest = self.__buffer
			self.__ypos -= 1	# update ypos
			self.__buffer = readline.get_history_item(self.__ypos)
			self.__xpos = len(self.__buffer)	# update cursor
			return True
		else: return False	# no past history to browse


	def __history_next(self):
		"""command to cause succeeding history to be displayed."""
		if self.__ypos < self.__ypos_index_latest():
			self.__ypos += 1
			# load latest
			if self.__is_latest(): self.__buffer = self.__latest
			else: self.__buffer = readline.get_history_item(self.__ypos)
			self.__xpos = len(self.__buffer)	# update cursor
			return True
		else: return False	# no forward history to browse


	# CURSOR ##############################################################
	def __cursor_prev(self):
		"""command to cause the cursor to move to the left."""
		if self.__xpos > 0:
			self.__xpos -= 1	# update xpos
			return True
		else: return False


	def __cursor_next(self):
		"""command to cause the cursor to move to the right."""
		if self.__xpos < len(self.__buffer):
			self.__xpos += 1	# update xpos
			return True
		else: return False


	def __cursor_del(self):
		"""command to delete a char. from before the current xpos."""
		# this condition check if we can do a <delete>
		if self.__xpos < len(self.__buffer):
			self.__buffer = self.__buffer[0:self.__xpos] + self.__buffer[self.__xpos+1:]
			# xpos *doesn't* get updated for this one.
			self.__history_replace()	# update
			return True
		else: return False


	def __cursor_bs(self):
		"""command to delete a char. from the current xpos."""
		# NOTE: we used to check the size of buffer to
		# ensure there actually were characters to del
		# but since we can assume buffer size via xpos
		# it makes more sense to just check that which
		# also lets you know if you can delete at left
		if self.__xpos > 0:
			# del the char from the buffer at xpos
			self.__buffer = self.__buffer[0:self.__xpos-1] + self.__buffer[self.__xpos:]
			self.__xpos -= 1
			self.__history_replace()	# update
			return True
		else: return False


	def __cursor_add(self, o):
		"""command to insert a char. from the current xpos."""
		try:
			if not chr(o) in self.good: return False
		except ValueError, e:
			return False
		# add the character to the buffer at xpos
		self.__buffer = self.__buffer[0:self.__xpos] + chr(o) + self.__buffer[self.__xpos:]
		self.__xpos += 1	# increment the position
		self.__history_replace()		# update
		return True


if __name__ == '__main__':
	# example
	mainloop = gobject.MainLoop()
	name = os.path.splitext(os.path.basename(__file__))[0]
	obj = gterminal(name)
	gobject.once_add_seconds(5, obj.msg, 'this is an example message.')
	try:
		mainloop.run()
	except KeyboardInterrupt, e:
		print '^C'

